Module: Puma

Defined in:
lib/puma.rb,
lib/puma/cli.rb,
lib/puma/dsl.rb,
lib/puma/util.rb,
lib/puma/const.rb,
lib/puma/binder.rb,
lib/puma/client.rb,
lib/puma/detect.rb,
lib/puma/events.rb,
lib/puma/plugin.rb,
lib/puma/runner.rb,
lib/puma/server.rb,
lib/puma/single.rb,
lib/puma/cluster.rb,
lib/puma/minissl.rb,
lib/puma/null_io.rb,
lib/puma/reactor.rb,
lib/puma/launcher.rb,
lib/puma/app/status.rb,
lib/puma/state_file.rb,
lib/puma/tcp_logger.rb,
lib/puma/control_cli.rb,
lib/puma/thread_pool.rb,
lib/puma/commonlogger.rb,
lib/puma/rack/builder.rb,
lib/puma/configuration.rb,
lib/puma/jruby_restart.rb,
lib/puma/minissl/context_builder.rb,
ext/puma_http11/mini_ssl.c,
ext/puma_http11/puma_http11.c

Defined Under Namespace

Modules: App, ConfigDefault, Const, JRubyRestart, MiniSSL, Rack, Util Classes: Binder, CLI, Client, Cluster, CommonLogger, Configuration, ConnectionError, ControlCLI, DSL, Events, HttpParser, HttpParserError, IOBuffer, Launcher, NullIO, Plugin, PluginLoader, PluginRegistry, Reactor, Runner, Server, Single, StateFile, TCPLogger, ThreadPool, UnknownPlugin, UnsupportedOption, UserFileDefaultOptions

Constant Summary collapse

HTTP_STATUS_CODES =

Every standard HTTP code mapped to the appropriate message. Generated with: curl -s www.iana.org/assignments/http-status-codes/http-status-codes-1.csv | \

ruby -ne 'm = /^(\d{3}),(?!Unassigned|\(Unused\))([^,]+)/.match($_) and \
          puts "#{m[1]} => \x27#{m[2].strip}\x27,"'
{
  100 => 'Continue',
  101 => 'Switching Protocols',
  102 => 'Processing',
  200 => 'OK',
  201 => 'Created',
  202 => 'Accepted',
  203 => 'Non-Authoritative Information',
  204 => 'No Content',
  205 => 'Reset Content',
  206 => 'Partial Content',
  207 => 'Multi-Status',
  208 => 'Already Reported',
  226 => 'IM Used',
  300 => 'Multiple Choices',
  301 => 'Moved Permanently',
  302 => 'Found',
  303 => 'See Other',
  304 => 'Not Modified',
  305 => 'Use Proxy',
  307 => 'Temporary Redirect',
  308 => 'Permanent Redirect',
  400 => 'Bad Request',
  401 => 'Unauthorized',
  402 => 'Payment Required',
  403 => 'Forbidden',
  404 => 'Not Found',
  405 => 'Method Not Allowed',
  406 => 'Not Acceptable',
  407 => 'Proxy Authentication Required',
  408 => 'Request Timeout',
  409 => 'Conflict',
  410 => 'Gone',
  411 => 'Length Required',
  412 => 'Precondition Failed',
  413 => 'Payload Too Large',
  414 => 'URI Too Long',
  415 => 'Unsupported Media Type',
  416 => 'Range Not Satisfiable',
  417 => 'Expectation Failed',
  418 => 'I\'m A Teapot',
  421 => 'Misdirected Request',
  422 => 'Unprocessable Entity',
  423 => 'Locked',
  424 => 'Failed Dependency',
  426 => 'Upgrade Required',
  428 => 'Precondition Required',
  429 => 'Too Many Requests',
  431 => 'Request Header Fields Too Large',
  451 => 'Unavailable For Legal Reasons',
  500 => 'Internal Server Error',
  501 => 'Not Implemented',
  502 => 'Bad Gateway',
  503 => 'Service Unavailable',
  504 => 'Gateway Timeout',
  505 => 'HTTP Version Not Supported',
  506 => 'Variant Also Negotiates',
  507 => 'Insufficient Storage',
  508 => 'Loop Detected',
  510 => 'Not Extended',
  511 => 'Network Authentication Required'
}
STATUS_WITH_NO_ENTITY_BODY =

For some HTTP status codes the client only expects headers.

{
  204 => true,
  205 => true,
  304 => true
}
IS_JRUBY =
defined?(JRUBY_VERSION)
IS_WINDOWS =
RUBY_PLATFORM =~ /mswin|ming|cygwin/
Plugins =
PluginRegistry.new

Class Attribute Summary collapse

Class Method Summary collapse

Class Attribute Details

.cli_configObject

The CLI exports its Puma::Configuration object here to allow apps to pick it up. An app needs to use it conditionally though since it is not set if the app is launched via another mechanism than the CLI class.



18
19
20
# File 'lib/puma/cli.rb', line 18

def cli_config
  @cli_config
end

Class Method Details

.jruby?Boolean

Returns:

  • (Boolean)


6
7
8
# File 'lib/puma/detect.rb', line 6

def self.jruby?
  IS_JRUBY
end

.set_thread_name(name) ⇒ Object

Thread name is new in Ruby 2.3



27
28
29
30
# File 'lib/puma.rb', line 27

def self.set_thread_name(name)
  return unless Thread.current.respond_to?(:name=)
  Thread.current.name = "puma #{name}"
end

.statsObject



22
23
24
# File 'lib/puma.rb', line 22

def self.stats
  @get_stats.stats
end

.stats_object=(val) ⇒ Object



18
19
20
# File 'lib/puma.rb', line 18

def self.stats_object=(val)
  @get_stats = val
end

.windows?Boolean

Returns:

  • (Boolean)


12
13
14
# File 'lib/puma/detect.rb', line 12

def self.windows?
  IS_WINDOWS
end