Method: Puma::DSL#on_worker_boot
- Defined in:
- lib/puma/dsl.rb
#on_worker_boot ⇒ Object
Note:
Cluster mode only.
Code to run in a worker when it boots to setup the process before booting the app.
This can be called multiple times to add several hooks.
830 831 832 833 834 |
# File 'lib/puma/dsl.rb', line 830 def before_worker_boot(key = nil, &block) Puma.deprecate_method_change :on_worker_boot, __callee__, __method__ process_hook :before_worker_boot, key, block, cluster_only: true end |