Class: Puma::ControlCLI
- Inherits:
-
Object
- Object
- Puma::ControlCLI
- Defined in:
- lib/puma/control_cli.rb
Constant Summary collapse
- COMMANDS =
%w{halt restart phased-restart start stats status stop reload-worker-directory gc gc-stats}
Instance Method Summary collapse
-
#initialize(argv, stdout = STDOUT, stderr = STDERR) ⇒ ControlCLI
constructor
A new instance of ControlCLI.
- #message(msg) ⇒ Object
- #prepare_configuration ⇒ Object
- #run ⇒ Object
- #send_request ⇒ Object
- #send_signal ⇒ Object
Constructor Details
#initialize(argv, stdout = STDOUT, stderr = STDERR) ⇒ ControlCLI
Returns a new instance of ControlCLI.
16 17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32 33 34 35 36 37 38 39 40 41 42 43 44 45 46 47 48 49 50 51 52 53 54 55 56 57 58 59 60 61 62 63 64 65 66 67 68 69 70 71 72 73 74 75 76 77 78 79 80 81 82 83 84 85 86 87 88 89 90 91 92 93 94 95 96 97 98 99 100 101 102 103 104 105 106 107 108 109 110 111 112 113 114 115 |
# File 'lib/puma/control_cli.rb', line 16 def initialize(argv, stdout=STDOUT, stderr=STDERR) @state = nil @quiet = false @pidfile = nil @pid = nil @control_url = nil @control_auth_token = nil @config_file = nil @command = nil @environment = ENV['RACK_ENV'] @argv = argv.dup @stdout = stdout @stderr = stderr = {} opts = OptionParser.new do |o| o. = "Usage: pumactl (-p PID | -P pidfile | -S status_file | -C url -T token | -F config.rb) (#{COMMANDS.join("|")})" o.on "-S", "--state PATH", "Where the state file to use is" do |arg| @state = arg end o.on "-Q", "--quiet", "Not display messages" do |arg| @quiet = true end o.on "-P", "--pidfile PATH", "Pid file" do |arg| @pidfile = arg end o.on "-p", "--pid PID", "Pid" do |arg| @pid = arg.to_i end o.on "-C", "--control-url URL", "The bind url to use for the control server" do |arg| @control_url = arg end o.on "-T", "--control-token TOKEN", "The token to use as authentication for the control server" do |arg| @control_auth_token = arg end o.on "-F", "--config-file PATH", "Puma config script" do |arg| @config_file = arg end o.on "-e", "--environment ENVIRONMENT", "The environment to run the Rack app on (default development)" do |arg| @environment = arg end o.on_tail("-H", "--help", "Show this message") do @stdout.puts o exit end o.on_tail("-V", "--version", "Show version") do puts Const::PUMA_VERSION exit end end opts.order!(argv) { |a| opts.terminate a } opts.parse! @command = argv.shift unless @config_file == '-' environment = @environment || 'development' if @config_file.nil? @config_file = %W(config/puma/#{environment}.rb config/puma.rb).find do |f| File.exist?(f) end end if @config_file config = Puma::Configuration.new({ config_files: [@config_file] }, {}) config.load @state ||= config.[:state] @control_url ||= config.[:control_url] @control_auth_token ||= config.[:control_auth_token] @pidfile ||= config.[:pidfile] end end # check present of command unless @command raise "Available commands: #{COMMANDS.join(", ")}" end unless COMMANDS.include? @command raise "Invalid command: #{@command}" end rescue => e @stdout.puts e. exit 1 end |
Instance Method Details
#message(msg) ⇒ Object
117 118 119 |
# File 'lib/puma/control_cli.rb', line 117 def (msg) @stdout.puts msg unless @quiet end |
#prepare_configuration ⇒ Object
121 122 123 124 125 126 127 128 129 130 131 132 133 134 135 136 137 |
# File 'lib/puma/control_cli.rb', line 121 def prepare_configuration if @state unless File.exist? @state raise "State file not found: #{@state}" end sf = Puma::StateFile.new sf.load @state @control_url = sf.control_url @control_auth_token = sf.control_auth_token @pid = sf.pid elsif @pidfile # get pid from pid_file File.open(@pidfile) { |f| @pid = f.read.to_i } end end |
#run ⇒ Object
249 250 251 252 253 254 255 256 257 258 259 260 261 262 263 |
# File 'lib/puma/control_cli.rb', line 249 def run return start if @command == "start" prepare_configuration if Puma.windows? send_request else @control_url ? send_request : send_signal end rescue => e e. exit 1 end |
#send_request ⇒ Object
139 140 141 142 143 144 145 146 147 148 149 150 151 152 153 154 155 156 157 158 159 160 161 162 163 164 165 166 167 168 169 170 171 172 173 174 175 176 177 178 179 180 181 182 183 184 185 186 187 188 189 190 191 192 193 194 |
# File 'lib/puma/control_cli.rb', line 139 def send_request uri = URI.parse @control_url # create server object by scheme server = case uri.scheme when "ssl" require 'openssl' OpenSSL::SSL::SSLSocket.new( TCPSocket.new(uri.host, uri.port), OpenSSL::SSL::SSLContext.new ).tap(&:connect) when "tcp" TCPSocket.new uri.host, uri.port when "unix" UNIXSocket.new "#{uri.host}#{uri.path}" else raise "Invalid scheme: #{uri.scheme}" end if @command == "status" "Puma is started" else url = "/#{@command}" if @control_auth_token url = url + "?token=#{@control_auth_token}" end server << "GET #{url} HTTP/1.0\r\n\r\n" unless data = server.read raise "Server closed connection before responding" end response = data.split("\r\n") if response.empty? raise "Server sent empty response" end (@http,@code,) = response.first.split(" ",3) if @code == "403" raise "Unauthorized access to server (wrong auth token)" elsif @code == "404" raise "Command error: #{response.last}" elsif @code != "200" raise "Bad response from server: #{@code}" end "Command #{@command} sent success" response.last if @command == "stats" || @command == "gc-stats" end ensure server.close if server && !server.closed? end |
#send_signal ⇒ Object
196 197 198 199 200 201 202 203 204 205 206 207 208 209 210 211 212 213 214 215 216 217 218 219 220 221 222 223 224 225 226 227 228 229 230 231 232 233 234 235 236 237 238 239 240 241 242 243 244 245 246 247 |
# File 'lib/puma/control_cli.rb', line 196 def send_signal unless @pid raise "Neither pid nor control url available" end begin case @command when "restart" Process.kill "SIGUSR2", @pid when "halt" Process.kill "QUIT", @pid when "stop" Process.kill "SIGTERM", @pid when "stats" puts "Stats not available via pid only" return when "reload-worker-directory" puts "reload-worker-directory not available via pid only" return when "phased-restart" Process.kill "SIGUSR1", @pid when "status" begin Process.kill 0, @pid puts "Puma is started" rescue Errno::ESRCH raise "Puma is not running" end return else return end rescue SystemCallError if @command == "restart" start else raise "No pid '#{@pid}' found" end end "Command #{@command} sent success" end |