Class: Puma::Launcher
- Inherits:
-
Object
- Object
- Puma::Launcher
- Defined in:
- lib/puma/launcher.rb
Overview
Puma::Launcher is the single entry point for starting a Puma server based on user configuration. It is responsible for taking user supplied arguments and resolving them with configuration in config/puma.rb or ‘config/puma/<env>.rb`.
It is responsible for either launching a cluster of Puma workers or a single puma server.
Constant Summary collapse
- KEYS_NOT_TO_PERSIST_IN_STATE =
[ :logger, :lowlevel_error_handler, :before_worker_shutdown, :before_worker_boot, :before_worker_fork, :after_worker_boot, :before_fork, :on_restart ]
Instance Attribute Summary collapse
-
#binder ⇒ Object
readonly
Returns the value of attribute binder.
-
#config ⇒ Object
readonly
Returns the value of attribute config.
-
#events ⇒ Object
readonly
Returns the value of attribute events.
-
#options ⇒ Object
readonly
Returns the value of attribute options.
-
#restart_dir ⇒ Object
readonly
Returns the value of attribute restart_dir.
Instance Method Summary collapse
- #close_binder_listeners ⇒ Object
-
#connected_port ⇒ Object
Return which tcp port the launcher is using, if it’s using TCP.
-
#delete_pidfile ⇒ Object
Delete the configured pidfile.
-
#halt ⇒ Object
Begin async shutdown of the server.
-
#initialize(conf, launcher_args = {}) ⇒ Launcher
constructor
Returns an instance of Launcher.
-
#phased_restart ⇒ Object
Begin a phased restart if supported.
-
#restart ⇒ Object
Begin async restart of the server.
- #restart_args ⇒ Object
-
#run ⇒ Object
Run the server.
-
#stats ⇒ Object
Return stats about the server.
-
#stop ⇒ Object
Begin async shutdown of the server gracefully.
-
#write_state ⇒ Object
Write a state file that can be used by pumactl to control the server.
Constructor Details
#initialize(conf, launcher_args = {}) ⇒ Launcher
Returns an instance of Launcher
conf A Puma::Configuration object indicating how to run the server.
launcher_args A Hash that currently has one required key :events, this is expected to hold an object similar to an Puma::Events.stdio, this object will be responsible for broadcasting Puma’s internal state to a logging destination. An optional key :argv can be supplied, this should be an array of strings, these arguments are re-used when restarting the puma server.
Examples:
conf = Puma::Configuration.new do |user_config|
user_config.threads 1, 10
user_config.app do |env|
[200, {}, ["hello world"]]
end
end
Puma::Launcher.new(conf, events: Puma::Events.stdio).run
43 44 45 46 47 48 49 50 51 52 53 54 55 56 57 58 59 60 61 62 63 64 65 66 67 68 69 70 71 72 73 74 75 76 77 78 79 80 81 82 83 84 85 86 87 88 89 90 91 92 93 |
# File 'lib/puma/launcher.rb', line 43 def initialize(conf, launcher_args={}) @runner = nil @events = launcher_args[:events] || Events::DEFAULT @argv = launcher_args[:argv] || [] @original_argv = @argv.dup @config = conf @binder = Binder.new(@events) @binder.import_from_env @environment = conf.environment # Advertise the Configuration Puma.cli_config = @config if defined?(Puma.cli_config) @config.load = @config. @config.clamp @events.formatter = Events::PidFormatter.new if clustered? @events.formatter = [:log_formatter] if [:log_formatter] generate_restart_data if clustered? && !Process.respond_to?(:fork) unsupported "worker mode not supported on #{RUBY_ENGINE} on this platform" end if [:daemon] && Puma.windows? unsupported 'daemon mode not supported on Windows' end Dir.chdir(@restart_dir) prune_bundler if prune_bundler? @environment = [:environment] if [:environment] set_rack_environment if clustered? [:logger] = @events @runner = Cluster.new(self, @events) else @runner = Single.new(self, @events) end Puma.stats_object = @runner @status = :run end |
Instance Attribute Details
#binder ⇒ Object (readonly)
Returns the value of attribute binder.
95 96 97 |
# File 'lib/puma/launcher.rb', line 95 def binder @binder end |
#config ⇒ Object (readonly)
Returns the value of attribute config.
95 96 97 |
# File 'lib/puma/launcher.rb', line 95 def config @config end |
#events ⇒ Object (readonly)
Returns the value of attribute events.
95 96 97 |
# File 'lib/puma/launcher.rb', line 95 def events @events end |
#options ⇒ Object (readonly)
Returns the value of attribute options.
95 96 97 |
# File 'lib/puma/launcher.rb', line 95 def end |
#restart_dir ⇒ Object (readonly)
Returns the value of attribute restart_dir.
95 96 97 |
# File 'lib/puma/launcher.rb', line 95 def restart_dir @restart_dir end |
Instance Method Details
#close_binder_listeners ⇒ Object
204 205 206 |
# File 'lib/puma/launcher.rb', line 204 def close_binder_listeners @binder.close_listeners end |
#connected_port ⇒ Object
Return which tcp port the launcher is using, if it’s using TCP
191 192 193 |
# File 'lib/puma/launcher.rb', line 191 def connected_port @binder.connected_port end |
#delete_pidfile ⇒ Object
Delete the configured pidfile
121 122 123 124 |
# File 'lib/puma/launcher.rb', line 121 def delete_pidfile path = [:pidfile] File.unlink(path) if path && File.exist?(path) end |
#halt ⇒ Object
Begin async shutdown of the server
127 128 129 130 |
# File 'lib/puma/launcher.rb', line 127 def halt @status = :halt @runner.halt end |
#phased_restart ⇒ Object
Begin a phased restart if supported
145 146 147 148 149 150 151 |
# File 'lib/puma/launcher.rb', line 145 def phased_restart unless @runner.respond_to?(:phased_restart) and @runner.phased_restart log "* phased-restart called but not available, restarting normally." return restart end true end |
#restart ⇒ Object
Begin async restart of the server
139 140 141 142 |
# File 'lib/puma/launcher.rb', line 139 def restart @status = :restart @runner.restart end |
#restart_args ⇒ Object
195 196 197 198 199 200 201 202 |
# File 'lib/puma/launcher.rb', line 195 def restart_args cmd = [:restart_cmd] if cmd cmd.split(' ') + @original_argv else @restart_argv end end |
#run ⇒ Object
Run the server. This blocks until the server is stopped
154 155 156 157 158 159 160 161 162 163 164 165 166 167 168 169 170 171 172 173 174 175 176 177 178 179 180 181 182 183 184 185 186 187 188 |
# File 'lib/puma/launcher.rb', line 154 def run previous_env = if defined?(Bundler) env = Bundler::ORIGINAL_ENV.dup # add -rbundler/setup so we load from Gemfile when restarting bundle = "-rbundler/setup" env["RUBYOPT"] = [env["RUBYOPT"], bundle].join(" ").lstrip unless env["RUBYOPT"].to_s.include?(bundle) env else ENV.to_h end @config.clamp @config.plugins.fire_starts self setup_signals set_process_title @runner.run case @status when :halt log "* Stopping immediately!" when :run, :stop graceful_stop when :restart log "* Restarting..." ENV.replace(previous_env) @runner.before_restart restart! when :exit # nothing end @binder.close_unix_paths end |
#stats ⇒ Object
Return stats about the server
98 99 100 |
# File 'lib/puma/launcher.rb', line 98 def stats @runner.stats end |
#stop ⇒ Object
Begin async shutdown of the server gracefully
133 134 135 136 |
# File 'lib/puma/launcher.rb', line 133 def stop @status = :stop @runner.stop end |
#write_state ⇒ Object
Write a state file that can be used by pumactl to control the server
104 105 106 107 108 109 110 111 112 113 114 115 116 117 118 |
# File 'lib/puma/launcher.rb', line 104 def write_state write_pid path = [:state] return unless path require 'puma/state_file' sf = StateFile.new sf.pid = Process.pid sf.control_url = [:control_url] sf.control_auth_token = [:control_auth_token] sf.save path end |