Class: Puma::Server
- Inherits:
-
Object
- Object
- Puma::Server
- Extended by:
- Forwardable
- Includes:
- Const
- Defined in:
- lib/puma/server.rb
Overview
The HTTP Server itself. Serves out a single Rack app.
This class is used by the Puma::Single and Puma::Cluster classes to generate one or more Puma::Server instances capable of handling requests. Each Puma process will contain one Puma::Server instance.
The Puma::Server instance pulls requests from the socket, adds them to a Puma::Reactor where they get eventually passed to a Puma::ThreadPool.
Each Puma::Server will have one reactor and one thread pool.
Constant Summary collapse
- UNPACK_TCP_STATE_FROM_TCP_INFO =
"C".freeze
- ThreadLocalKey =
:puma_server
Constants included from Const
Const::CGI_VER, Const::CHUNKED, Const::CHUNK_SIZE, Const::CLOSE, Const::CLOSE_CHUNKED, Const::CODE_NAME, Const::COLON, Const::CONNECTION_CLOSE, Const::CONNECTION_KEEP_ALIVE, Const::CONTENT_LENGTH, Const::CONTENT_LENGTH2, Const::CONTENT_LENGTH_S, Const::CONTINUE, Const::EARLY_HINTS, Const::ERROR_RESPONSE, Const::FAST_TRACK_KA_TIMEOUT, Const::FIRST_DATA_TIMEOUT, Const::GATEWAY_INTERFACE, Const::HALT_COMMAND, Const::HEAD, Const::HIJACK, Const::HIJACK_IO, Const::HIJACK_P, Const::HTTP, Const::HTTPS, Const::HTTPS_KEY, Const::HTTP_10_200, Const::HTTP_11, Const::HTTP_11_100, Const::HTTP_11_200, Const::HTTP_CONNECTION, Const::HTTP_EXPECT, Const::HTTP_HOST, Const::HTTP_INJECTION_REGEX, Const::HTTP_VERSION, Const::HTTP_X_FORWARDED_FOR, Const::HTTP_X_FORWARDED_PROTO, Const::HTTP_X_FORWARDED_SCHEME, Const::HTTP_X_FORWARDED_SSL, Const::KEEP_ALIVE, Const::LINE_END, Const::LOCALHOST, Const::LOCALHOST_ADDR, Const::LOCALHOST_IP, Const::MAX_BODY, Const::MAX_FAST_INLINE, Const::MAX_HEADER, Const::NEWLINE, Const::PATH_INFO, Const::PERSISTENT_TIMEOUT, Const::PORT_443, Const::PORT_80, Const::PUMA_CONFIG, Const::PUMA_PEERCERT, Const::PUMA_SERVER_STRING, Const::PUMA_SOCKET, Const::PUMA_TMP_BASE, Const::PUMA_VERSION, Const::QUERY_STRING, Const::RACK_AFTER_REPLY, Const::RACK_INPUT, Const::RACK_URL_SCHEME, Const::REMOTE_ADDR, Const::REQUEST_METHOD, Const::REQUEST_PATH, Const::REQUEST_URI, Const::RESTART_COMMAND, Const::SERVER_NAME, Const::SERVER_PORT, Const::SERVER_PROTOCOL, Const::SERVER_SOFTWARE, Const::STOP_COMMAND, Const::TRANSFER_ENCODING, Const::TRANSFER_ENCODING2, Const::TRANSFER_ENCODING_CHUNKED, Const::WORKER_CHECK_INTERVAL, Const::WRITE_TIMEOUT
Instance Attribute Summary collapse
-
#app ⇒ Object
Returns the value of attribute app.
-
#auto_trim_time ⇒ Object
Returns the value of attribute auto_trim_time.
-
#binder ⇒ Object
Returns the value of attribute binder.
-
#early_hints ⇒ Object
Returns the value of attribute early_hints.
-
#events ⇒ Object
readonly
Returns the value of attribute events.
-
#first_data_timeout ⇒ Object
Returns the value of attribute first_data_timeout.
-
#leak_stack_on_error ⇒ Object
Returns the value of attribute leak_stack_on_error.
-
#max_threads ⇒ Object
Returns the value of attribute max_threads.
-
#min_threads ⇒ Object
Returns the value of attribute min_threads.
-
#persistent_timeout ⇒ Object
Returns the value of attribute persistent_timeout.
-
#reaping_time ⇒ Object
Returns the value of attribute reaping_time.
-
#thread ⇒ Object
readonly
Returns the value of attribute thread.
Class Method Summary collapse
Instance Method Summary collapse
- #backlog ⇒ Object
- #begin_restart ⇒ Object
- #closed_socket?(socket) ⇒ Boolean
-
#cork_socket(socket) ⇒ Object
6 == Socket::IPPROTO_TCP 3 == TCP_CORK 1/0 == turn on/off.
- #default_server_port(env) ⇒ Object
-
#graceful_shutdown ⇒ Object
Wait for all outstanding requests to finish.
- #halt(sync = false) ⇒ Object
-
#handle_check ⇒ Object
:nodoc:.
-
#handle_request(req, lines) ⇒ Object
Takes the request
req, invokes the Rack application to construct the response and writes it back toreq.io. - #handle_servers ⇒ Object
- #handle_servers_lopez_mode ⇒ Object
- #inherit_binder(bind) ⇒ Object
-
#initialize(app, events = Events.stdio, options = {}) ⇒ Server
constructor
Create a server for the rack app
app. -
#lowlevel_error(e, env) ⇒ Object
A fallback rack response if @app raises as exception.
-
#normalize_env(env, client) ⇒ Object
Given a Hash
envfor the request read fromclient, add and fixup keys to comply with Rack’s env guidelines. -
#pool_capacity ⇒ Object
This number represents the number of requests that the server is capable of taking right now.
-
#process_client(client, buffer) ⇒ Object
Given a connection on
client, handle the incoming requests. -
#read_body(env, client, body, cl) ⇒ Object
Given the request
envfromclientand the partial bodybodyplus a potential Content-Length valuecl, finish reading the body and return it. -
#run(background = true) ⇒ Object
Runs the server.
-
#run_lopez_mode(background = true) ⇒ Object
Lopez Mode == raw tcp apps.
- #running ⇒ Object
- #shutting_down? ⇒ Boolean
-
#stop(sync = false) ⇒ Object
Stops the acceptor thread and then causes the worker threads to finish off the request queue before finally exiting.
- #tcp_mode! ⇒ Object
- #uncork_socket(socket) ⇒ Object
Constructor Details
#initialize(app, events = Events.stdio, options = {}) ⇒ Server
Create a server for the rack app app.
events is an object which will be called when certain error events occur to be handled. See Puma::Events for the list of current methods to implement.
Server#run returns a thread that you can join on to wait for the server to do its work.
56 57 58 59 60 61 62 63 64 65 66 67 68 69 70 71 72 73 74 75 76 77 78 79 80 81 82 83 84 85 86 87 88 |
# File 'lib/puma/server.rb', line 56 def initialize(app, events=Events.stdio, ={}) @app = app @events = events @check, @notify = Puma::Util.pipe @status = :stop @min_threads = 0 @max_threads = 16 @auto_trim_time = 30 @reaping_time = 1 @thread = nil @thread_pool = nil @early_hints = nil @persistent_timeout = .fetch(:persistent_timeout, PERSISTENT_TIMEOUT) @first_data_timeout = .fetch(:first_data_timeout, FIRST_DATA_TIMEOUT) @binder = Binder.new(events) @leak_stack_on_error = true @options = @queue_requests = [:queue_requests].nil? ? true : [:queue_requests] ENV['RACK_ENV'] ||= "development" @mode = :http @precheck_closing = true end |
Instance Attribute Details
#app ⇒ Object
Returns the value of attribute app.
39 40 41 |
# File 'lib/puma/server.rb', line 39 def app @app end |
#auto_trim_time ⇒ Object
Returns the value of attribute auto_trim_time.
44 45 46 |
# File 'lib/puma/server.rb', line 44 def auto_trim_time @auto_trim_time end |
#binder ⇒ Object
Returns the value of attribute binder.
90 91 92 |
# File 'lib/puma/server.rb', line 90 def binder @binder end |
#early_hints ⇒ Object
Returns the value of attribute early_hints.
90 91 92 |
# File 'lib/puma/server.rb', line 90 def early_hints @early_hints end |
#events ⇒ Object (readonly)
Returns the value of attribute events.
38 39 40 |
# File 'lib/puma/server.rb', line 38 def events @events end |
#first_data_timeout ⇒ Object
Returns the value of attribute first_data_timeout.
46 47 48 |
# File 'lib/puma/server.rb', line 46 def first_data_timeout @first_data_timeout end |
#leak_stack_on_error ⇒ Object
Returns the value of attribute leak_stack_on_error.
90 91 92 |
# File 'lib/puma/server.rb', line 90 def leak_stack_on_error @leak_stack_on_error end |
#max_threads ⇒ Object
Returns the value of attribute max_threads.
42 43 44 |
# File 'lib/puma/server.rb', line 42 def max_threads @max_threads end |
#min_threads ⇒ Object
Returns the value of attribute min_threads.
41 42 43 |
# File 'lib/puma/server.rb', line 41 def min_threads @min_threads end |
#persistent_timeout ⇒ Object
Returns the value of attribute persistent_timeout.
43 44 45 |
# File 'lib/puma/server.rb', line 43 def persistent_timeout @persistent_timeout end |
#reaping_time ⇒ Object
Returns the value of attribute reaping_time.
45 46 47 |
# File 'lib/puma/server.rb', line 45 def reaping_time @reaping_time end |
#thread ⇒ Object (readonly)
Returns the value of attribute thread.
37 38 39 |
# File 'lib/puma/server.rb', line 37 def thread @thread end |
Class Method Details
.current ⇒ Object
1070 1071 1072 |
# File 'lib/puma/server.rb', line 1070 def self.current Thread.current[ThreadLocalKey] end |
Instance Method Details
#backlog ⇒ Object
155 156 157 |
# File 'lib/puma/server.rb', line 155 def backlog @thread_pool and @thread_pool.backlog end |
#begin_restart ⇒ Object
1043 1044 1045 |
# File 'lib/puma/server.rb', line 1043 def begin_restart notify_safely(RESTART_COMMAND) end |
#closed_socket?(socket) ⇒ Boolean
127 128 129 130 131 132 133 134 135 136 137 138 139 140 141 142 |
# File 'lib/puma/server.rb', line 127 def closed_socket?(socket) return false unless socket.kind_of? TCPSocket return false unless @precheck_closing begin tcp_info = socket.getsockopt(Socket::SOL_TCP, Socket::TCP_INFO) rescue IOError, SystemCallError Thread.current.purge_interrupt_queue if Thread.current.respond_to? :purge_interrupt_queue @precheck_closing = false false else state = tcp_info.unpack(UNPACK_TCP_STATE_FROM_TCP_INFO)[0] # TIME_WAIT: 6, CLOSE: 7, CLOSE_WAIT: 8, LAST_ACK: 9, CLOSING: 11 (state >= 6 && state <= 9) || state == 11 end end |
#cork_socket(socket) ⇒ Object
6 == Socket::IPPROTO_TCP 3 == TCP_CORK 1/0 == turn on/off
111 112 113 114 115 116 117 |
# File 'lib/puma/server.rb', line 111 def cork_socket(socket) begin socket.setsockopt(6, 3, 1) if socket.kind_of? TCPSocket rescue IOError, SystemCallError Thread.current.purge_interrupt_queue if Thread.current.respond_to? :purge_interrupt_queue end end |
#default_server_port(env) ⇒ Object
607 608 609 610 611 612 613 |
# File 'lib/puma/server.rb', line 607 def default_server_port(env) if ['on', HTTPS].include?(env[HTTPS_KEY]) || env[HTTP_X_FORWARDED_PROTO].to_s[0...5] == HTTPS || env[HTTP_X_FORWARDED_SCHEME] == HTTPS || env[HTTP_X_FORWARDED_SSL] == "on" PORT_443 else PORT_80 end end |
#graceful_shutdown ⇒ Object
Wait for all outstanding requests to finish.
962 963 964 965 966 967 968 969 970 971 972 973 974 975 976 977 978 979 980 981 982 983 984 985 986 987 988 989 990 991 992 993 994 995 996 997 998 999 1000 1001 1002 1003 1004 1005 1006 1007 1008 1009 1010 1011 |
# File 'lib/puma/server.rb', line 962 def graceful_shutdown if @options[:shutdown_debug] threads = Thread.list total = threads.size pid = Process.pid $stdout.syswrite "#{pid}: === Begin thread backtrace dump ===\n" threads.each_with_index do |t,i| $stdout.syswrite "#{pid}: Thread #{i+1}/#{total}: #{t.inspect}\n" $stdout.syswrite "#{pid}: #{t.backtrace.join("\n#{pid}: ")}\n\n" end $stdout.syswrite "#{pid}: === End thread backtrace dump ===\n" end if @options[:drain_on_shutdown] count = 0 while true ios = IO.select @binder.ios, nil, nil, 0 break unless ios ios.first.each do |sock| begin if io = sock.accept_nonblock count += 1 client = Client.new io, @binder.env(sock) @thread_pool << client end rescue SystemCallError end end end @events.debug "Drained #{count} additional connections." end if @status != :restart @binder.close end if @thread_pool if timeout = @options[:force_shutdown_after] @thread_pool.shutdown timeout.to_i else @thread_pool.shutdown end end end |
#halt(sync = false) ⇒ Object
1038 1039 1040 1041 |
# File 'lib/puma/server.rb', line 1038 def halt(sync=false) notify_safely(HALT_COMMAND) @thread.join if @thread && sync end |
#handle_check ⇒ Object
:nodoc:
439 440 441 442 443 444 445 446 447 448 449 450 451 452 453 454 455 |
# File 'lib/puma/server.rb', line 439 def handle_check cmd = @check.read(1) case cmd when STOP_COMMAND @status = :stop return true when HALT_COMMAND @status = :halt return true when RESTART_COMMAND @status = :restart return true end return false end |
#handle_request(req, lines) ⇒ Object
Takes the request req, invokes the Rack application to construct the response and writes it back to req.io.
The second parameter lines is a IO-like object unique to this thread. This is normally an instance of Puma::IOBuffer.
It’ll return false when the connection is closed, this doesn’t mean that the response wasn’t successful.
It’ll return :async if the connection remains open but will be handled elsewhere, i.e. the connection has been hijacked by the Rack application.
Finally, it’ll return true on keep-alive connections.
628 629 630 631 632 633 634 635 636 637 638 639 640 641 642 643 644 645 646 647 648 649 650 651 652 653 654 655 656 657 658 659 660 661 662 663 664 665 666 667 668 669 670 671 672 673 674 675 676 677 678 679 680 681 682 683 684 685 686 687 688 689 690 691 692 693 694 695 696 697 698 699 700 701 702 703 704 705 706 707 708 709 710 711 712 713 714 715 716 717 718 719 720 721 722 723 724 725 726 727 728 729 730 731 732 733 734 735 736 737 738 739 740 741 742 743 744 745 746 747 748 749 750 751 752 753 754 755 756 757 758 759 760 761 762 763 764 765 766 767 768 769 770 771 772 773 774 775 776 777 778 779 780 781 782 783 784 785 786 787 788 789 790 791 792 793 794 795 796 797 798 799 800 801 802 803 804 805 806 807 808 809 810 811 812 813 814 815 816 817 818 819 820 821 822 823 824 825 826 827 828 829 830 831 832 833 834 835 836 837 838 839 840 841 842 843 844 845 846 847 848 849 850 851 852 853 854 855 856 857 858 859 860 861 862 863 864 865 866 867 868 869 870 871 872 873 874 875 876 877 878 879 880 881 |
# File 'lib/puma/server.rb', line 628 def handle_request(req, lines) env = req.env client = req.io return false if closed_socket?(client) normalize_env env, req env[PUMA_SOCKET] = client if env[HTTPS_KEY] && client.peercert env[PUMA_PEERCERT] = client.peercert end env[HIJACK_P] = true env[HIJACK] = req body = req.body head = env[REQUEST_METHOD] == HEAD env[RACK_INPUT] = body env[RACK_URL_SCHEME] = default_server_port(env) == PORT_443 ? HTTPS : HTTP if @early_hints env[EARLY_HINTS] = lambda { |headers| begin fast_write client, "HTTP/1.1 103 Early Hints\r\n".freeze headers.each_pair do |k, vs| if vs.respond_to?(:to_s) && !vs.to_s.empty? vs.to_s.split(NEWLINE).each do |v| next if possible_header_injection?(v) fast_write client, "#{k}: #{v}\r\n" end else fast_write client, "#{k}: #{vs}\r\n" end end fast_write client, "\r\n".freeze rescue ConnectionError # noop, if we lost the socket we just won't send the early hints end } end # Fixup any headers with , in the name to have _ now. We emit # headers with , in them during the parse phase to avoid ambiguity # with the - to _ conversion for critical headers. But here for # compatibility, we'll convert them back. This code is written to # avoid allocation in the common case (ie there are no headers # with , in their names), that's why it has the extra conditionals. to_delete = nil to_add = nil env.each do |k,v| if k.start_with?("HTTP_") and k.include?(",") and k != "HTTP_TRANSFER,ENCODING" if to_delete to_delete << k else to_delete = [k] end unless to_add to_add = {} end to_add[k.gsub(",", "_")] = v end end if to_delete to_delete.each { |k| env.delete(k) } env.merge! to_add end # A rack extension. If the app writes #call'ables to this # array, we will invoke them when the request is done. # after_reply = env[RACK_AFTER_REPLY] = [] begin begin status, headers, res_body = @app.call(env) return :async if req.hijacked status = status.to_i if status == -1 unless headers.empty? and res_body == [] raise "async response must have empty headers and body" end return :async end rescue ThreadPool::ForceShutdown => e @events.log "Detected force shutdown of a thread, returning 503" @events.unknown_error self, e, "Rack app" status = 503 headers = {} res_body = ["Request was internally terminated early\n"] rescue Exception => e @events.unknown_error self, e, "Rack app", env status, headers, res_body = lowlevel_error(e, env) end content_length = nil no_body = head if res_body.kind_of? Array and res_body.size == 1 content_length = res_body[0].bytesize end cork_socket client line_ending = LINE_END colon = COLON http_11 = if env[HTTP_VERSION] == HTTP_11 allow_chunked = true keep_alive = env.fetch(HTTP_CONNECTION, "").downcase != CLOSE include_keepalive_header = false # An optimization. The most common response is 200, so we can # reply with the proper 200 status without having to compute # the response header. # if status == 200 lines << HTTP_11_200 else lines.append "HTTP/1.1 ", status.to_s, " ", fetch_status_code(status), line_ending no_body ||= status < 200 || STATUS_WITH_NO_ENTITY_BODY[status] end true else allow_chunked = false keep_alive = env.fetch(HTTP_CONNECTION, "").downcase == KEEP_ALIVE include_keepalive_header = keep_alive # Same optimization as above for HTTP/1.1 # if status == 200 lines << HTTP_10_200 else lines.append "HTTP/1.0 ", status.to_s, " ", fetch_status_code(status), line_ending no_body ||= status < 200 || STATUS_WITH_NO_ENTITY_BODY[status] end false end response_hijack = nil headers.each do |k, vs| case k.downcase when CONTENT_LENGTH2 next if possible_header_injection?(vs) content_length = vs next when TRANSFER_ENCODING allow_chunked = false content_length = nil when HIJACK response_hijack = vs next end if vs.respond_to?(:to_s) && !vs.to_s.empty? vs.to_s.split(NEWLINE).each do |v| next if possible_header_injection?(v) lines.append k, colon, v, line_ending end else lines.append k, colon, line_ending end end if include_keepalive_header lines << CONNECTION_KEEP_ALIVE elsif http_11 && !keep_alive lines << CONNECTION_CLOSE end if no_body if content_length and status != 204 lines.append CONTENT_LENGTH_S, content_length.to_s, line_ending end lines << line_ending fast_write client, lines.to_s return keep_alive end if content_length lines.append CONTENT_LENGTH_S, content_length.to_s, line_ending chunked = false elsif !response_hijack and allow_chunked lines << TRANSFER_ENCODING_CHUNKED chunked = true end lines << line_ending fast_write client, lines.to_s if response_hijack response_hijack.call client return :async end begin res_body.each do |part| next if part.bytesize.zero? if chunked fast_write client, part.bytesize.to_s(16) fast_write client, line_ending fast_write client, part fast_write client, line_ending else fast_write client, part end client.flush end if chunked fast_write client, CLOSE_CHUNKED client.flush end rescue SystemCallError, IOError raise ConnectionError, "Connection error detected during write" end ensure uncork_socket client body.close req.tempfile.unlink if req.tempfile res_body.close if res_body.respond_to? :close after_reply.each { |o| o.call } end return keep_alive end |
#handle_servers ⇒ Object
364 365 366 367 368 369 370 371 372 373 374 375 376 377 378 379 380 381 382 383 384 385 386 387 388 389 390 391 392 393 394 395 396 397 398 399 400 401 402 403 404 405 406 407 408 409 410 411 412 413 414 415 416 417 418 419 420 421 422 423 424 425 426 427 428 429 430 431 432 433 434 435 436 |
# File 'lib/puma/server.rb', line 364 def handle_servers begin check = @check sockets = [check] + @binder.ios pool = @thread_pool queue_requests = @queue_requests remote_addr_value = nil remote_addr_header = nil case @options[:remote_address] when :value remote_addr_value = @options[:remote_address_value] when :header remote_addr_header = @options[:remote_address_header] end while @status == :run begin ios = IO.select sockets ios.first.each do |sock| if sock == check break if handle_check else begin if io = sock.accept_nonblock client = Client.new io, @binder.env(sock) if remote_addr_value client.peerip = remote_addr_value elsif remote_addr_header client.remote_addr_header = remote_addr_header end pool << client busy_threads = pool.wait_until_not_full if busy_threads == 0 @options[:out_of_band].each(&:call) if @options[:out_of_band] end end rescue SystemCallError # nothing rescue Errno::ECONNABORTED # client closed the socket even before accept begin io.close rescue Thread.current.purge_interrupt_queue if Thread.current.respond_to? :purge_interrupt_queue end end end end rescue Object => e @events.unknown_error self, e, "Listen loop" end end @events.fire :state, @status graceful_shutdown if @status == :stop || @status == :restart if queue_requests @reactor.clear! @reactor.shutdown end rescue Exception => e STDERR.puts "Exception handling servers: #{e.} (#{e.class})" STDERR.puts e.backtrace ensure @check.close @notify.close end @events.fire :state, :done end |
#handle_servers_lopez_mode ⇒ Object
217 218 219 220 221 222 223 224 225 226 227 228 229 230 231 232 233 234 235 236 237 238 239 240 241 242 243 244 245 246 247 248 249 250 251 252 253 254 255 256 257 258 259 260 261 262 263 264 265 266 267 268 269 270 271 272 273 274 275 |
# File 'lib/puma/server.rb', line 217 def handle_servers_lopez_mode begin check = @check sockets = [check] + @binder.ios pool = @thread_pool while @status == :run begin ios = IO.select sockets ios.first.each do |sock| if sock == check break if handle_check else begin if io = sock.accept_nonblock client = Client.new io, nil pool << client end rescue SystemCallError # nothing rescue Errno::ECONNABORTED # client closed the socket even before accept begin io.close rescue Thread.current.purge_interrupt_queue if Thread.current.respond_to? :purge_interrupt_queue end end end end rescue Object => e @events.unknown_error self, e, "Listen loop" end end @events.fire :state, @status graceful_shutdown if @status == :stop || @status == :restart rescue Exception => e STDERR.puts "Exception handling servers: #{e.} (#{e.class})" STDERR.puts e.backtrace ensure begin @check.close rescue Thread.current.purge_interrupt_queue if Thread.current.respond_to? :purge_interrupt_queue end # Prevent can't modify frozen IOError (RuntimeError) begin @notify.close rescue IOError # no biggy end end @events.fire :state, :done end |
#inherit_binder(bind) ⇒ Object
94 95 96 |
# File 'lib/puma/server.rb', line 94 def inherit_binder(bind) @binder = bind end |
#lowlevel_error(e, env) ⇒ Object
A fallback rack response if @app raises as exception.
944 945 946 947 948 949 950 951 952 953 954 955 956 957 958 |
# File 'lib/puma/server.rb', line 944 def lowlevel_error(e, env) if handler = @options[:lowlevel_error_handler] if handler.arity == 1 return handler.call(e) else return handler.call(e, env) end end if @leak_stack_on_error [500, {}, ["Puma caught this error: #{e.} (#{e.class})\n#{e.backtrace.join("\n")}"]] else [500, {}, ["An unhandled lowlevel error occurred. The application logs may have details.\n"]] end end |
#normalize_env(env, client) ⇒ Object
Given a Hash env for the request read from client, add and fixup keys to comply with Rack’s env guidelines.
554 555 556 557 558 559 560 561 562 563 564 565 566 567 568 569 570 571 572 573 574 575 576 577 578 579 580 581 582 583 584 585 586 587 588 589 590 591 592 593 594 595 596 597 598 599 600 601 602 603 604 605 |
# File 'lib/puma/server.rb', line 554 def normalize_env(env, client) if host = env[HTTP_HOST] if colon = host.index(":") env[SERVER_NAME] = host[0, colon] env[SERVER_PORT] = host[colon+1, host.bytesize] else env[SERVER_NAME] = host env[SERVER_PORT] = default_server_port(env) end else env[SERVER_NAME] = LOCALHOST env[SERVER_PORT] = default_server_port(env) end unless env[REQUEST_PATH] # it might be a dumbass full host request header uri = URI.parse(env[REQUEST_URI]) env[REQUEST_PATH] = uri.path raise "No REQUEST PATH" unless env[REQUEST_PATH] # A nil env value will cause a LintError (and fatal errors elsewhere), # so only set the env value if there actually is a value. env[QUERY_STRING] = uri.query if uri.query end env[PATH_INFO] = env[REQUEST_PATH] # From http://www.ietf.org/rfc/rfc3875 : # "Script authors should be aware that the REMOTE_ADDR and # REMOTE_HOST meta-variables (see sections 4.1.8 and 4.1.9) # may not identify the ultimate source of the request. # They identify the client for the immediate request to the # server; that client may be a proxy, gateway, or other # intermediary acting on behalf of the actual source client." # unless env.key?(REMOTE_ADDR) begin addr = client.peerip rescue Errno::ENOTCONN # Client disconnects can result in an inability to get the # peeraddr from the socket; default to localhost. addr = LOCALHOST_IP end # Set unix socket addrs to localhost addr = LOCALHOST_IP if addr.empty? env[REMOTE_ADDR] = addr end end |
#pool_capacity ⇒ Object
This number represents the number of requests that the server is capable of taking right now.
For example if the number is 5 then it means there are 5 threads sitting idle ready to take a request. If one request comes in, then the value would be 4 until it finishes processing.
171 172 173 |
# File 'lib/puma/server.rb', line 171 def pool_capacity @thread_pool and @thread_pool.pool_capacity end |
#process_client(client, buffer) ⇒ Object
Given a connection on client, handle the incoming requests.
This method support HTTP Keep-Alive so it may, depending on if the client indicates that it supports keep alive, wait for another request before returning.
463 464 465 466 467 468 469 470 471 472 473 474 475 476 477 478 479 480 481 482 483 484 485 486 487 488 489 490 491 492 493 494 495 496 497 498 499 500 501 502 503 504 505 506 507 508 509 510 511 512 513 514 515 516 517 518 519 520 521 522 523 524 525 526 527 528 529 530 531 532 533 534 535 536 537 538 539 540 541 542 543 544 545 546 547 548 549 |
# File 'lib/puma/server.rb', line 463 def process_client(client, buffer) begin clean_thread_locals = @options[:clean_thread_locals] close_socket = true requests = 0 while true case handle_request(client, buffer) when false return when :async close_socket = false return when true return unless @queue_requests buffer.reset ThreadPool.clean_thread_locals if clean_thread_locals requests += 1 check_for_more_data = @status == :run if requests >= MAX_FAST_INLINE # This will mean that reset will only try to use the data it already # has buffered and won't try to read more data. What this means is that # every client, independent of their request speed, gets treated like a slow # one once every MAX_FAST_INLINE requests. check_for_more_data = false end unless client.reset(check_for_more_data) close_socket = false client.set_timeout @persistent_timeout @reactor.add client return end end end # The client disconnected while we were reading data rescue ConnectionError # Swallow them. The ensure tries to close +client+ down # SSL handshake error rescue MiniSSL::SSLError => e lowlevel_error(e, client.env) ssl_socket = client.io addr = ssl_socket.peeraddr.last cert = ssl_socket.peercert close_socket = true @events.ssl_error self, addr, cert, e # The client doesn't know HTTP well rescue HttpParserError => e lowlevel_error(e, client.env) client.write_error(400) @events.parse_error self, client.env, e # Server error rescue StandardError => e lowlevel_error(e, client.env) client.write_error(500) @events.unknown_error self, e, "Read" ensure buffer.reset begin client.close if close_socket rescue IOError, SystemCallError Thread.current.purge_interrupt_queue if Thread.current.respond_to? :purge_interrupt_queue # Already closed rescue StandardError => e @events.unknown_error self, e, "Client" end end end |
#read_body(env, client, body, cl) ⇒ Object
Given the request env from client and the partial body body plus a potential Content-Length value cl, finish reading the body and return it.
If the body is larger than MAX_BODY, a Tempfile object is used for the body, otherwise a StringIO is used.
895 896 897 898 899 900 901 902 903 904 905 906 907 908 909 910 911 912 913 914 915 916 917 918 919 920 921 922 923 924 925 926 927 928 929 930 931 932 933 934 935 936 937 938 939 940 |
# File 'lib/puma/server.rb', line 895 def read_body(env, client, body, cl) content_length = cl.to_i remain = content_length - body.bytesize return StringIO.new(body) if remain <= 0 # Use a Tempfile if there is a lot of data left if remain > MAX_BODY stream = Tempfile.new(Const::PUMA_TMP_BASE) stream.binmode else # The body[0,0] trick is to get an empty string in the same # encoding as body. stream = StringIO.new body[0,0] end stream.write body # Read an odd sized chunk so we can read even sized ones # after this chunk = client.readpartial(remain % CHUNK_SIZE) # No chunk means a closed socket unless chunk stream.close return nil end remain -= stream.write(chunk) # Raed the rest of the chunks while remain > 0 chunk = client.readpartial(CHUNK_SIZE) unless chunk stream.close return nil end remain -= stream.write(chunk) end stream.rewind return stream end |
#run(background = true) ⇒ Object
Runs the server.
If background is true (the default) then a thread is spun up in the background to handle requests. Otherwise requests are handled synchronously.
282 283 284 285 286 287 288 289 290 291 292 293 294 295 296 297 298 299 300 301 302 303 304 305 306 307 308 309 310 311 312 313 314 315 316 317 318 319 320 321 322 323 324 325 326 327 328 329 330 331 332 333 334 335 336 337 338 339 340 341 342 343 344 345 346 347 348 349 350 351 352 353 354 355 356 357 358 359 360 361 362 |
# File 'lib/puma/server.rb', line 282 def run(background=true) BasicSocket.do_not_reverse_lookup = true @events.fire :state, :booting @status = :run if @mode == :tcp return run_lopez_mode(background) end queue_requests = @queue_requests @thread_pool = ThreadPool.new(@min_threads, @max_threads, IOBuffer) do |client, buffer| # Advertise this server into the thread Thread.current[ThreadLocalKey] = self process_now = false begin if queue_requests process_now = client.eagerly_finish else client.finish process_now = true end rescue MiniSSL::SSLError => e ssl_socket = client.io addr = ssl_socket.peeraddr.last cert = ssl_socket.peercert client.close @events.ssl_error self, addr, cert, e rescue HttpParserError => e client.write_error(400) client.close @events.parse_error self, client.env, e rescue ConnectionError, EOFError client.close else if process_now process_client client, buffer else client.set_timeout @first_data_timeout @reactor.add client end end end @thread_pool.clean_thread_locals = @options[:clean_thread_locals] if queue_requests @reactor = Reactor.new self, @thread_pool @reactor.run_in_thread end if @reaping_time @thread_pool.auto_reap!(@reaping_time) end if @auto_trim_time @thread_pool.auto_trim!(@auto_trim_time) end @events.fire :state, :running if background @thread = Thread.new do Puma.set_thread_name "server" handle_servers end return @thread else handle_servers end end |
#run_lopez_mode(background = true) ⇒ Object
Lopez Mode == raw tcp apps
177 178 179 180 181 182 183 184 185 186 187 188 189 190 191 192 193 194 195 196 197 198 199 200 201 202 203 204 205 206 207 208 209 210 211 212 213 214 215 |
# File 'lib/puma/server.rb', line 177 def run_lopez_mode(background=true) @thread_pool = ThreadPool.new(@min_threads, @max_threads, Hash) do |client, tl| io = client.to_io addr = io.peeraddr.last if addr.empty? # Set unix socket addrs to localhost addr = "127.0.0.1:0" else addr = "#{addr}:#{io.peeraddr[1]}" end env = { 'thread' => tl, REMOTE_ADDR => addr } begin @app.call env, client.to_io rescue Object => e STDERR.puts "! Detected exception at toplevel: #{e.} (#{e.class})" STDERR.puts e.backtrace end client.close unless env['detach'] end @events.fire :state, :running if background @thread = Thread.new do Puma.set_thread_name "server" handle_servers_lopez_mode end return @thread else handle_servers_lopez_mode end end |
#running ⇒ Object
159 160 161 |
# File 'lib/puma/server.rb', line 159 def running @thread_pool and @thread_pool.spawned end |
#shutting_down? ⇒ Boolean
1074 1075 1076 |
# File 'lib/puma/server.rb', line 1074 def shutting_down? @status == :stop || @status == :restart end |
#stop(sync = false) ⇒ Object
Stops the acceptor thread and then causes the worker threads to finish off the request queue before finally exiting.
1033 1034 1035 1036 |
# File 'lib/puma/server.rb', line 1033 def stop(sync=false) notify_safely(STOP_COMMAND) @thread.join if @thread && sync end |
#tcp_mode! ⇒ Object
98 99 100 |
# File 'lib/puma/server.rb', line 98 def tcp_mode! @mode = :tcp end |
#uncork_socket(socket) ⇒ Object
119 120 121 122 123 124 125 |
# File 'lib/puma/server.rb', line 119 def uncork_socket(socket) begin socket.setsockopt(6, 3, 0) if socket.kind_of? TCPSocket rescue IOError, SystemCallError Thread.current.purge_interrupt_queue if Thread.current.respond_to? :purge_interrupt_queue end end |